#include <png.h>

#include <cmath>
using std::floor;

#include <string>
using std::string;

#include <iostream>
using std::cerr;

class bitmap
{
private:
	int width;
	int height;
	const int bit_depth;
	const int default_compression;
	unsigned char **buffer;
	double filegamma_;
	double screengamma_;
	
public:
	bitmap(int w, int h, double b) : width(w), height(h), bit_depth(16), default_compression(6)
	{
		filegamma_ = 0.6;
		
		int backgroundcolour = int(b*65535);
		if (width < 0 or height < 0)
		{
			cerr << " ERROR: Constructor called with negative height or width. Setting width and height to 1." << std::endl;
			width = 1;
			height = 1;
		}
		
		if (backgroundcolour > 65535)
		{
			cerr << " WARNING: Constructor called with background colour greater than 65535. Setting to 65535."<<std::endl;
			backgroundcolour = 65535;
		}
		
		if (backgroundcolour < 0)
		{
			cerr << " WARNING: Constructor called with background colour lower than 0. Setting to 0."<<std::endl;
			backgroundcolour = 0;
		}
		
		screengamma_ = 2.2;
		
		//buffer = (png_bytepp)malloc(height * sizeof(png_bytep));
		buffer = new png_bytep[height];
		if (buffer == NULL)
		{
			cerr << " ERROR:  Not able to allocate memory for image." << std::endl;
		}
		
		for (h = 0; h < height; h++)
		{
			//buffer[h] = (png_bytep)malloc(6*width * sizeof(png_byte));
			buffer[h] = new png_byte[6*width];
			if (buffer[h] == NULL)
			{
				cerr << " ERROR:  Not able to allocate memory for image." << std::endl;
			}
		}
		
		if (buffer == NULL)
		{
			cerr << " ERROR:  Not able to allocate memory for image." << std::endl;
		}
		
		for (w = 0; w < width; w++)
		{
			for (h = 0; h < height; h++)
			{
				buffer[h][6*w] = (char) floor(((double)backgroundcolour)/256);
				buffer[h][6*w+1] = (char)(backgroundcolour%256);
				buffer[h][6*w+2] = (char) floor(((double)backgroundcolour)/256);
				buffer[h][6*w+3] = (char)(backgroundcolour%256);
				buffer[h][6*w+4] = (char) floor(((double)backgroundcolour)/256);
				buffer[h][6*w+5] = (char)(backgroundcolour%256);
			}
		}
	}
	
	~bitmap()
	{
		for (int h = 0; h < height; ++h) 
			delete [] buffer[h];
		delete [] buffer;
	}

	void write(string filename)
	{
		FILE            *fp;
		png_structp     png_ptr;
		png_infop       info_ptr;
		
		fp = fopen(filename.c_str(), "wb");
		if ( fp == NULL)
		{
			cerr << "ERROR: Error creating file (fopen() returned NULL pointer)." << std::endl;
			return;
		}
		
		png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
		info_ptr = png_create_info_struct(png_ptr);
		png_init_io(png_ptr, fp);
		png_set_compression_level(png_ptr, default_compression);
		
		png_set_IHDR(png_ptr, info_ptr, width, height,
					 bit_depth, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
					 PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
		
		if (filegamma_ < 1.0e-1)
		{
			filegamma_ = 0.5;
		}
		
		png_set_gAMA(png_ptr, info_ptr, filegamma_);
		
		time_t          gmt;
		png_time        mod_time;
		png_text        text_ptr[1];
		time(&gmt);
		png_convert_from_time_t(&mod_time, gmt);
		png_set_tIME(png_ptr, info_ptr, &mod_time);
		png_set_text(png_ptr, info_ptr, text_ptr, 0);
		png_write_info(png_ptr, info_ptr);
		png_write_image(png_ptr, buffer);
		png_write_end(png_ptr, info_ptr);
		png_destroy_write_struct(&png_ptr, &info_ptr);
		fclose(fp);
	}
	
	void set_pixel(int x, int y, int red, int green, int blue)
	{
		++x;
		++y;
		if (red > 65535) red = 65535;
		if (green > 65535) green = 65535;
		if (blue > 65535) blue = 65535;
		
		if (red < 0) red = 0;
		if (green < 0) green = 0;
		if (blue < 0) blue = 0;
		
		if (bit_depth == 16)
		{
			if ( y <= height and y > 0 and x > 0 and x <= width )
			{
				buffer[height-y][6*x-6] = (char) floor(((double)red)/256);
				buffer[height-y][6*x-5] = (char)(red%256);
				buffer[height-y][6*x-4] = (char) floor(((double)green)/256);
				buffer[height-y][6*x-3] = (char)(green%256);
				buffer[height-y][6*x-2] = (char) floor(((double)blue)/256);
				buffer[height-y][6*x-1] = (char)(blue%256);
			};
		}
		
		if ((bit_depth == 8))
		{
			if ( (y<height+1) and (y>0) and (x>0) and (x<width+1) )
			{
				buffer[height-y][3*x-3] = (char)(floor(((double)red)/257.0));
				buffer[height-y][3*x-2] = (char)(floor(((double)green)/257.0));
				buffer[height-y][3*x-1] = (char)(floor(((double)blue)/257.0));
				
			};
		}
	}
	
	void set_pixel(int x, int y, double red, double green, double blue)
	{
		set_pixel(x,y,int(red*65535),int(green*65535),int(blue*65535));
	}
	
};

